implementation module GameLibrary

import StdEnv, StdIO
import StdGameDef, StdGame, StdGSt
import StdProcess
import notes

OBJ_AUTOINIT  :==  0


//:: GameStateFunction :== *(GSt DefaultGameState) -> *(GSt DefaultGameState)

defaultGame :: .(Game DefaultGameState) [.GameAttribute DefaultGameState] !*World -> .World
defaultGame gamedef options world
    # world = startGame gamedef (defaultInitialGameState gamedef.levels)
                     (options ++ [ScreenSize {w=640,h=480}, ColorDepth 16]) world
    = world

::  DefaultGameState
    = { curlevel    :: Int
      , maxlevel    :: Int
      , quit        :: Bool

      }

defaultInitialGameState levels = { curlevel = 0
                                 , maxlevel = length levels
                                 , quit = False
                                 
                                 }


defaultGameDef :: [Level (GSt DefaultGameState)] -> (Game DefaultGameState)
defaultGameDef levels =
    { levels = levels
    , quitlevel = accGSt quitFunction
    , nextlevel = accGSt nextLevelFunction
    , textitems = accGSt textItemsFunction
    }


/* if the quit function returns true, the game engine quit the level */

quitFunction :: DefaultGameState -> (Bool, DefaultGameState)
quitFunction gst
    = (gst.quit, {gst & quit = False})


/* quit the level */

quitlevel :: !*(GameObjectState a *(GSt DefaultGameState)) -> *GameObjectState a *(GSt DefaultGameState);
quitlevel objst=:{st, or, gs}
    #   gs = appGSt setQuit gs
    =   {objst & gs = gs}
where
    setQuit :: DefaultGameState -> DefaultGameState
    setQuit gst = {gst & quit = True}


/* function that returns the next level to run, 0 = end game */

nextLevelFunction :: DefaultGameState -> (Int, DefaultGameState)
nextLevelFunction gst =: {curlevel, maxlevel}
    = (nextLevel, {gst & curlevel = nextLevel})
where
    nextLevel = if (curlevel + 1 > maxlevel) 0 (curlevel + 1)


/* function that returns text to be displayed */

textItemsFunction :: DefaultGameState -> ([GameText], DefaultGameState)
textItemsFunction gst
    = ([], gst)

defaultGameLevel :: !(Maybe .Colour) ![.Layer] !.BoundMap !.[GameObject (*GSt .a)] (IdFun (*GSt .a)) [SoundSample] -> (.Level (*GSt .a)) 
defaultGameLevel fillcolor layers boundmap objects f sndsamples
    = { blankScreen
      & boundmap     = boundmap
      , layers       = layers
      , objects      = [(mainObject f)] ++ objects 
      , leveloptions = { blankScreen.leveloptions
                       & fillbackground = fillcolor
                       , debugscroll = True
                       }
      , soundsamples = sndsamples
      }
where
    mainObject :: (*(GSt .a) -> *(GSt .a)) -> (GameObject *(GSt .a))
    mainObject f
        # obj = defaultGameObject OBJ_AUTOINIT {w=1,h=1} Void
        # obj = {obj & init = (newInit {w=1,h=1} Void)}
        = obj
    where
        newInit size state subcode pos time gs
            # (objrec, gs) = defaultObjectRec subcode pos size time gs
            # gs = f gs
            = {st=state, or={objrec & active = False}, gs=gs}


emptyMap :: BoundMap
emptyMap = { map = [{0}]
           , blocksize = {w = 1600, h = 1200}
           , objstart  = 1
           , startobjx = 1
           , startobjy = 1
           }

defaultBoundMap :: !Size !Size ![(Point2, Int)] -> BoundMap
defaultBoundMap blocksize mapsize objinst
    = { emptyMap
      & map = [ListToArray [(f x y objinst) \\ x <- [1..mapsize.w]] \\ y <- [1..mapsize.h]]
      , blocksize = blocksize
      }
where
    f x y [] = 0
    f x y [(p, i):xs]
        | (p.x == x) && (p.y == y) = i * 256    // shl 8
        | otherwise                = f x y xs

ListToArray :: ![Int] -> {#Int}
ListToArray xs = {x \\ x <- xs}

ArrayToList :: !{#Int} -> [Int]
ArrayToList xs = [x \\ x <-: xs]

newObject :: ObjectCode Size [Sprite] (Point2 -> GameObjectRec -> GameObjectRec) -> (GameObject (GSt a))
newObject objtype size spritelist initfun
  = { defaultGameObject objtype size Void
    & sprites = spritelist
    , init    = (newinit size Void)
    }
where
    newinit size state subcode pos time gs
        # (objrec, gs) = defaultObjectRec subcode pos size time gs
        # objrec = initfun pos objrec
        = {st = state, or = objrec, gs = gs}

createGameObject :: !ObjectCode !Point2 !(GSt .gs) -> (GSt .gs)
createGameObject oc p gs
    # (_, gs) = createNewGameObject oc 0 p gs
    = gs

noInstances :: *(GSt .gs) -> *(GSt .gs)
noInstances x = x

defaultTransparentSprite :: !String !Size !Point2 -> Sprite
defaultTransparentSprite file size point
    # spr = defaultSprite file size
    # spr = {spr & bitmap = {spr.bitmap & transparent = Just point}}
    = spr

defaultSprite :: !String !Size -> Sprite
defaultSprite file size
    = {bitmap = newbitmap, sequence = [(1,10000)], loop = True}
where
    newbitmap = { bitmapname = file
                , unitsize = size
                , dimensions = (1,1)
                , transparent = Nothing
                }

sound :: !Int !*(GSt .gs) -> *(GSt .gs)
sound index gs
    # (_, gs) = playSoundSample index MAX_VOLUME PAN_CENTER (getnotefreq 72) 0 gs
    = gs
